

import {all as L} from './infra/log.js';  // logger
import {all as API} from './infra/api.js';  // server rest api
import {all as DB} from './infra/db.js';  // indexeddb
import {all as LANG} from './infra/lang.js';  // languages
import {all as S} from './domain/search.js';  // search
import {all as R} from './domain/rsrc.js';  // resource 
import {all as T} from './domain/task.js';  // task
import {all as I} from './icon.js';  // action icon
import {all as U} from './utils.js';  // utilities


/* ==========================================================
 * Use Cases
 */

const favToggle = async (url) => {
  L.log('favToggle', [url])
  if (! U.testUrl(url)) {
    throw new Error('invalid url: ' + url)
  }
  url = U.urlToASCII(url)
  return await DB.transaction('rw', ['rsrcs', 'fgtasks'], async () => {
    const found = await DB.findRsrcByUrl(url)
    if (! found) {
      const rsrc = R.create(url)
      const task = T.createFavAdd([url], true, 0)
      await DB.storeTask(task)
      R.setCurrentTask(rsrc, task.id, true)
      await DB.storeRsrc(rsrc)
      return true
    } else {
      const curTask = R.currentTask(found)
      if (curTask) {
        await DB.deleteTask(curTask[0], curTask[1])
        // rsrcを消すのでR.clearCurrentTask()は不要
      }
      await DB.deleteRsrc(found.id)  // rsrcだけ消す。indexはタスクで消す。
      const task = T.createFavDel([url], [found.id], true, 0)
      await DB.storeTask(task)
      return false
    }
  })
}

const flip = async (task) => {
  L.log('flip', [task])
  return await DB.transaction('rw', ['props', 'fgtasks', 'rsrcs'], async () => {
    const batch = await DB.findProp(task.batchName)
    if (! batch || batch.serial != task.batchSerial) {
      // The batch which this task belongs had been deleted. Do nothing
      return
    }

    let urls = []
    for (let url of task.urls) {
      if (! U.testUrl(url)) continue
      url = U.urlToASCII(url)
      const found = await DB.findRsrcByUrl(url)
      if (found) continue
      urls.push(url)
    }

    const newTask = T.createFavAdd(urls, true, 0)
    await DB.storeTask(newTask)
    for (let url of urls) {
      const rsrc = R.create(url)
      R.setCurrentTask(rsrc, newTask.id, true)
      await DB.storeRsrc(rsrc)
    }
    T.incrCurrentStep(batch)
    await DB.storeProp(batch.name, batch)
  })
}

// favAdd()は冪等
const favAdd = async (task) => {
  L.log('favAdd', [task])
  const langs = await LANG.getAcceptLanguages()
  // addFav APIは冪等
  // Dexieのトランザクションの中で別由来のPromiseを使ってはならない。http://bit.ly/2kdckMn
  await API.addFav(task.urls, langs)
  return await DB.transaction('rw', ['rsrcs', 'fgtasks'], async () => {
    for (let i = 0; i < task.urls.length; i++) {
      const rsrc = await DB.findRsrcByUrl(task.urls[i])
      const curTask = R.currentTask(rsrc)
      if (curTask) {
        await DB.deleteTask(curTask[0], curTask[1])
      }
      const nextTask = T.createCapture(task.urls[i], langs, task.prior, 2500)
      await DB.storeTask(nextTask)
      R.setCurrentTask(rsrc, nextTask.id, nextTask.prior)
      await DB.storeRsrc(rsrc)
    }
  })
}

// captureは冪等
const capture = async (task) => {
  L.log('capture', [task])
  // captureは冪等
  // Dexieのトランザクションの中で別由来のPromiseを使ってはならない。http://bit.ly/2kdckMn
  const gotRsrc = await API.capture(task.url)
  const ready = task.langs.some(l => gotRsrc.langs.hasOwnProperty(l))
  return await DB.transaction('rw', ['rsrcs', 'indice', 'fgtasks'], async () => {
    const rsrc = await DB.findRsrcByUrl(task.url)
    if (ready) {
      if (R.isChanged(rsrc, gotRsrc)) {
        const indice = R.capture(rsrc, gotRsrc, task.langs)
        await DB.deleteIndice(rsrc.id)
        await DB.storeIndice(indice)
      }
      R.clearCurrentTask(rsrc)
      await DB.storeRsrc(rsrc)
    } else {
      const newTask = T.reschedule(task)
      await DB.storeTask(newTask)
      R.setCurrentTask(rsrc, newTask.id, newTask.prior)
      await DB.storeRsrc(rsrc)
    }
  })
}

// favDelは冪等
const favDel = async (task) => {
  L.log('favDel', [task])
  // delfav APIは冪等
  // Dexieのトランザクションの中で別由来のPromiseを使ってはならない。http://bit.ly/2kdckMn
  await API.delFav(task.urls)
  return await DB.transaction('rw', ['indice', 'fgtasks'], async () => {
    for (let rsrcId of task.rsrcIds) {
      await DB.deleteIndice(rsrcId)
    }
  })
}

const favTest = async (url) => {
  L.log('favTest', [url])
  url = U.urlToASCII(url)
  const rsrc = await DB.findRsrcByUrl(url)
  return (rsrc != null)
}

const favList = async (page, options) => {
  // https://web.dev/file-system-access/
  if (! options) options = {}
  L.log('favList', [page, options])
  const RSRC_PER_PAGE = 50
  const langs = await LANG.getAcceptLanguages()
  const lang = langs[0]
  const offset = (page - 1) * RSRC_PER_PAGE
  const limit = RSRC_PER_PAGE
  return await DB.transaction('r', ['rsrcs'], async () => {
    const {rsrcs, total} = await DB.sliceRsrcs(offset, limit)  // non-atomic DB operation
    const hits = rsrcs.map(r => {
      const hit = R.toHit(r, lang)
      if (options.urlAsIs) {
        return hit
      } else {
        return {...hit, url:U.urlToUnicode(hit.url)}
      }
    })
    const lastPage = total ? Math.ceil(total / RSRC_PER_PAGE) : null
    return {hits, total, lastPage}
  })
}

const search = async (keyword, options) => {
  if (! options) options = {}
  L.log('search', [keyword, options])
  const query = S.parse(keyword)
  const langs = await LANG.getAcceptLanguages()
  const lang = langs[0]

  return await DB.transaction('r', ['indice', 'rsrcs'], async () => {
    const rsrcIds = await S.search(query, lang, DB, false)
    const rsrcs = []
    for (let rsrcId of rsrcIds) {
      const rsrc = await DB.findRsrc(rsrcId)
      if (rsrc) rsrcs.push(rsrc)  // favの削除中はrsrcがnullの場合もある。
    }
    const hits = rsrcs.map(rsrc => {
      const hit = R.toHit(rsrc, lang)
      if (options.urlAsIs) {
        return hit
      } else {
        return {...hit, url:U.urlToUnicode(hit.url)}
      }
    })
    L.log('search-hits', hits)
    return hits
  })
}

// -> batch|null
const batchAdd = async (name) => {
  L.log('batchAdd', [name])
  return await DB.transaction('rw', ['props'], async () => {
    const found = await DB.findProp(name)
    if (found) return null
    const batch = T.createBatch(name)
    await DB.storeProp(name, batch)
    return batch
  })
}

// -> batch|null
const batchGet = async (name) => {
  //L.log('batchGet', [name])
  const found = await DB.findProp(name)
  return found ? found : null
}

// -> void
const batchDel = async (name) => {
  L.log('batchDel', [name])
  return await DB.deleteProp(name)
}

// batchPrepareだけはserialを照合する。
// batchPrepareでは、どのバッチ処理にデータを流し込むかを指定する必要があるため。
// batchGetやbatchDelは、どの種類のバッチ処理を操作対象とするかを指定しているに過ぎない。
// URLのASCII化はここでは行わない。flipで行う。
// -> void
const batchPrepare = async (name, serial, urls) => {
  L.log('batchPrepare', [name, serial, urls])
  return await DB.transaction('rw', ['props', 'bgtasks'], async () => {
    const batch = await DB.findProp(name)
    if (! batch || batch.serial != serial) {
      throw new Error(`no such batch ${name}, ${serial}`)
    }

    const task = T.createFlip(urls, name, batch.serial, false, 0)
    await DB.storeTask(task)
    T.incrTotalStep(batch)
    await DB.storeProp(name, batch)
  })
}

const dumplog = async () => {
  const logs = L.dump()
  const content = JSON.stringify(logs)
  return Promise.resolve(content)
}

const xParse = async (keyword) => {
  return Promise.resolve(S.parse(keyword))
}

const xSearch = async (keyword) => {
  const query = S.parse(keyword)
  const langs = await LANG.getAcceptLanguages()
  const lang = langs[0]

  const entries = await S.search(query, lang, DB, true)
  const results = []
  for (let entry of entries) {
    const rsrc = await DB.findRsrc(entry.rsrc)
    if (! rsrc) continue;  // favの削除中はrsrcがnullの場合もある。
    const title = rsrc.titles.hasOwnProperty(lang) ? rsrc.titles[lang] : 'unknown'
    const url = rsrc.url
    const score = entry.score
    const id = rsrc.id
    results.push({id, score, title, url})
  }

  return results
}

const xIndice = async (termOrId) => {
  const cook = (indice) => {
    return indice.map(idx => `${idx.terms[0][0]} (${idx.score}); id=${idx.id}, rsrc=${idx.rsrc}`)
  }
  const langs = await LANG.getAcceptLanguages()
  const lang = langs[0]

  if (termOrId) {
    if (typeof termOrId == 'number' || termOrId === ""+(+termOrId)) {
      return cook(await DB.sliceIndexByRsrc(termOrId))
    } else {
      return cook(await DB.sliceIndex(term, lang))
    }
  } else {
    return cook(await DB.listIndice())
  }
}

const xProp = async (name) => {
  if (name) {
    return await DB.findProp(name)
  } else {
    return await DB.listProps()
  }
}

const xTask = async (prior) => {
  return await DB.listTasks(prior)
}

const xRsrc = async (urlOrId) => {
  if (urlOrId) {
    if (("" + urlOrId).startsWith('http')) {
      // is url
      const url = U.urlToASCII(urlOrId)
      return await DB.findRsrcByUrl(url)
    } else {
      // rsrcId
      return await DB.findRsrc(urlOrId)
    }
  } else {
    return await DB.listRsrcs()
  }
}

const xDumplog = async () => {
  const logs = L.dump()
  return Promise.resolve(logs.map(ent => {
    const when = (new Date(ent.when)).toJSON()
    return {...ent, when}
  }))
}

const xDeleteTasks = async (prior, taskIds) => {
  for (let taskId of taskIds) {
    await DB.deleteTask(taskId, prior)
  }
}

const xDeleteRsrcs = async (rsrcIds) => {
  for (let rsrcId of rsrcIds) {
    await DB.deleteRsrc(rsrcId)
  }
}

const xDeleteProp = async (name) => {
  await DB.deleteProp(name)
}

const xClear = async () => {
  await DB.clear()
}


/* ==========================================================
 * Initializing & Bindings
 */

const getConfig = (mode) => {
  mode = 'production'
  if (mode == 'development') {
    return {
      dbname: 'favalot-devel9', 
      baseUrl: 'http://localhost', 
      rsrcBaseUrl: 'https://storage.googleapis.com/favalot-rsrc-test/', 
      langs: ['ja', 'en'], 
      admin: true, 
      sendLogToConsole: true
    }
  } else {
    return {
      dbname: 'favalot', 
      baseUrl: 'https://api.favalot.com', 
      rsrcBaseUrl: 'https://storage.googleapis.com/favalot-rsrc/', 
      langs: ['ja', 'en'], 
      admin: false, 
      sendLogToConsole: false
    }
  }
}
const init = async () => {
  const info = await chrome.management.get(chrome.runtime.id)
  const {dbname, baseUrl, rsrcBaseUrl, langs, admin, sendLogToConsole} = getConfig(info.installType)
  L.init(sendLogToConsole)
  L.log('init', [info.installType])
  await LANG.init(langs)
  await DB.init(dbname)
  await API.init(baseUrl, rsrcBaseUrl, DB, L)

  const messageHandlers = {
    favToggle: {func:favToggle, args:["url"]}, 
    favTest: {func:favTest, args:["url"]}, 
    favList: {func:favList, args:["page", "options"]}, 
    search: {func:search, args:["keyword", "options"]}, 
    batchAdd: {func:batchAdd, args:["name"]}, 
    batchGet: {func:batchGet, args:["name"]}, 
    batchDel: {func:batchDel, args:["name"]}, 
    batchPrepare: {func:batchPrepare, args:["name", "serial", "urls"]}, 
    dumplog: {func:dumplog, args:[]}
  }
  const dispatch = (message, sender, sendResponse) => {
    // dispatchは関数コールかonMessageハンドラとしてのみ使われる。
    // なので、当拡張の内部からしか呼び出せず、その意味でセキュア。
    // 認証・認可は行わない。
    // dispatchから呼ばれる各ユースケース関数はサーバーへの通信を
    // 行わないので、サーバーと通信ができなくても当アプリは動く。
    // 例外はAPI.init()
    const handlerSpec = messageHandlers[message.type]
    if (! handlerSpec) {
      L.error('unknown message on dispatch', message.type)
      return
    }
    const func = handlerSpec.func
    const args = handlerSpec.args.map(name => message[name])
    return func(...args)
    .then((result) => {
      sendResponse({ok:true, value:result})
    })
    .catch((err) => {
      L.error(`on ${message.type}`, L.flattenError(err))
      sendResponse({ok:false, value:null})
    })
  }
  chrome.runtime.onMessage.addListener((message, sender, sendResponse) => {
    dispatch(message, sender, sendResponse)
    return true
  })
  I.init(async (message) => {
    let result = null
    const sendResponse = (result0) => {result = result0;}
    await dispatch(message, null, sendResponse)
    return result
  })

  const taskHandlers = {
    favAdd, 
    favDel, 
    capture, 
    flip
  }
  const dispatchCron = async (scheduledTime) => {
    // dispatchCronは内部コールまたはonAlarmハンドラとしてのみ使われる。
    // なのでセキュア。認証・認可は行わない。
    const limit = scheduledTime + 55*1000
    const iterationMax = 100
    let iterCount = 0
    let forceExit = false
    while (Date.now() < limit && ! forceExit && iterCount < iterationMax) {
      iterCount++
      let task = await DB.pickTask(true)
      if (! task) {
        task = await DB.pickTask(false)
      }
      if (! task) {
        await U.sleep(5000)
      } else {
        const handler = taskHandlers[task.type]
        if (! handler) {
          L.error('unknown task', task)
          await DB.deleteTask(task.id, task.prior)
        } else {
          await handler(task)
          .then(() => {
            DB.deleteTask(task.id, task.prior)
          })
          .catch((err) => {
            L.error('on ' + task.type, L.flattenError(err))
            T.putOff(task)
            return DB.storeTask(task)
          })
        }
        await U.sleep(200)
      }
    }
    if (forceExit) {
      L.log('dispatchCron-end', {cause:'forceExit', iterCount})
    } else if (iterCount >= iterationMax) {
      L.log('dispatchCron-end', {cause:'iterCount', iterCount})
    } else if (Date.now() >= limit) {
      L.log('dispatchCron-end', {cause:'timeLimit', iterCount})
    } else {
      L.log('dispatchCron-end', {cause:'unknown', iterCount})
    }
  }
  chrome.alarms.onAlarm.addListener((alarm) => dispatchCron(alarm.scheduledTime))
  chrome.alarms.create("", {periodInMinutes: 1})
  setTimeout(() => {
    dispatchCron(Date.now())
  }, 100)

  if (admin) {
    self.xfavalot = async (...args) => {
      switch (args[0]) {
        case 'indice': 
          return xIndice(args[1])
        case 'prop': 
          return xProp(args[1])
        case 'task': 
          return xTask(args[1])
        case 'rsrc': 
          return xRsrc(args[1])
        case 'parse': 
          return xParse(args[1])
        case 'search': 
          return xSearch(args[1])
        case 'delete-task': 
          return xDeleteTasks(args[1], args[2])
        case 'delete-rsrc': 
          return xDeleteRsrcs(args[1])
        case 'delete-prop': 
          return xDeleteProp(args[1])
        case 'clear': 
          return xClear()
        case 'dumplog': 
          return xDumplog()
        default: 
          console.log(
            'Usage: \n' + 
            '  xfavalot("indice", [termOrId:string|int])\n' +
            '  xfavalot("prop", [name:string])\n' + 
            '  xfavalot("task", prior:boolean)\n' + 
            '  xfavalot("parse", keyword:string)\n' + 
            '  xfavalot("search", keyword:string)\n' + 
            '  xfavalot("rsrc", [urlOrId:string|int])\n' + 
            '  xfavalot("delete-task", prior:boolean, taskIds:Array<number>)\n' + 
            '  xfavalot("delete-rsrc", rsrcIds:Array<number>)\n' + 
            '  xfavalot("delete-prop", name:string)\n' + 
            '  xfavalot("clean")\n' + 
            '  xfavalot("dumplog")'
          )
          return null
      }
    }
  }
}

// 起動段階でエラーが発生するなら稼働する意味は無い。エラーハンドリングせずに落ちる。
init()
