
/**
 * - ほとんどの場合リソースに対する表現は1つ。リソース表現を別レコードにする価値は少ないので、すべてrsrcに入れる
 * 
 * rsrc
 * - id: auto-generated primary key
 * - url: string  -- unique indexed
 * - currentTask: [key,prior]|null
 * - pickAt: number|null  -- indexed ピック時刻。作成時・更新タスク開始時・キャプチャ完了時に更新される
 * - titles: Record<string,number>  -- lang => title
 * 
 */
/**
 * index  -- 複数のrsrcをまとめない場合
 * - id: auto-generated primary key
 * - terms: Array<[term, lang]>  -- multi-entry index 検索時
 * - rsrc: key  -- index rsrc.id 削除時
 * - score: number
 */


// -> rsrc
export const create = (url) => {
  return {
    id: null, 
    currentTask: null, 
    pickAt: null, 
    url, 
    titles: {}, 
    etag: ""  // This etag never matches to server-generated etags.
  }
}

// -> void
export const setCurrentTask = (rsrc, taskId, prior) => {
  rsrc.currentTask = [taskId, prior]
}

// -> void
export const clearCurrentTask = (rsrc) => {
  rsrc.currentTask = null
}

// -> [id,prior]|null
export const currentTask = (rsrc) => {
  return rsrc.currentTask
}

export const isChanged = (rsrc, gotRsrc) => {
  return rsrc.etag != gotRsrc.etag
}

// -> index
const createIndex = (terms, rsrcId, score) => {
  return {
    id: null, 
    terms,  // Array<[term, lang]>
    rsrc: rsrcId, 
    score
  }
}

// -> index[]
export const capture = (rsrc, gotRsrc, langs) => {
  const titles = {}  // lang -> title
  const reprs = {}  // key -> {repr, langs}
  for (let lang of langs) {
    const key = gotRsrc.langs[lang]
    if (! key) continue
    const repr = gotRsrc.reprs[key]
    if (repr.status != 200) continue
    titles[lang] = repr.title
    if (! reprs[key]) {
      reprs[key] = {repr, langs:[lang]}
    } else {
      reprs[key].langs.push(lang)
    }
  }

  rsrc.titles = titles
  rsrc.pickAt = Math.floor(Date.now() / 1000)

  const indice = []
  for (let key in reprs) {
    // id, terms, rsrc, score;  terms = Array<[term, lang]>
    const {repr, langs} = reprs[key]
    for (let term in repr.terms) {
      const terms = langs.map(lang => [term, lang])
      const index = createIndex(terms, rsrc.id, repr.terms[term])
      indice.push(index)
    }
  }

  return indice
}

// -> {url:string, title:string}
export const toHit = (rsrc, lang) => {
  const title = rsrc.titles.hasOwnProperty(lang) ? rsrc.titles[lang] : 'unknown'
  return {url:rsrc.url, text:title}
}

export const all = {create, setCurrentTask, clearCurrentTask, currentTask, isChanged, capture, toHit}