
import {all as U} from '../utils.js';  // utilities

const reSp  = /\s+/y
const reLpar = /[\(（]/y
const reRpar = /[\)）]/y
const reOp  = /\+|AND|\||OR/y
const reUop = /-|@|\$|#|[a-z]+:/y
const reQuot = /"([^"$]*)"?/y
const reLit = /[^\s"\(\)（）]+/y

// -> token[]
// where token = {type, word, off}; type = UOP | OP | LPAR | RPAR | LIT | QUOT | EOP
const tokenize = (phrase, offset) => {
  const ts = []
  const len = phrase.length
  let ms
  let pos = 0
  let spAllowed = true

  while (pos < len) {
    if (spAllowed) {
      // skip spaces
      reSp.lastIndex = pos
      if (reSp.exec(phrase)) {
        pos = reSp.lastIndex
      }
      if (pos >= len) break
    }

    // UOP -- unary operator
    reUop.lastIndex = pos
    if (ms = reUop.exec(phrase)) {
      ts.push({type:'UOP', word:ms[0], pos:pos + offset})
      pos = reUop.lastIndex
      spAllowed = false
      continue
    }

    // OP -- infix operator
    reOp.lastIndex = pos
    if (ms = reOp.exec(phrase)) {
      const op = ms[0].toLowerCase() == 'and' ? '&' 
               : ms[0].toLowerCase() == '+' ? '&' 
               : ms[0].toLowerCase() == 'or' ? '|'
               : '|'
      ts.push({type:'OP', word:op, pos:pos + offset})
      pos = reOp.lastIndex
      spAllowed = true
      continue
    }

    // LIT -- literal
    reLit.lastIndex = pos
    if (ms = reLit.exec(phrase)) {
      const word = U.halfAlnum(U.kataKana(U.fullKana(ms[0]))).toLowerCase()
      ts.push({type:'LIT', word, pos:pos + offset})
      pos = reLit.lastIndex
      spAllowed = true
      continue
    }

    // LPAR -- left paren
    reLpar.lastIndex = pos
    if (reLpar.exec(phrase)) {
      ts.push({type:'LPAR', word:ms[0], pos:pos + offset})
      pos = reLpar.lastIndex
      spAllowed = true
      continue
    }

    // RPAR -- right paren
    reRpar.lastIndex = pos
    if (reRpar.exec(phrase)) {
      ts.push({type:'RPAR', word:ms[0], pos:pos + offset})
      pos = reRpar.lastIndex
      spAllowed = true
      continue
    }

    // QUOT -- quoted
    reQuot.lastIndex = pos
    if (ms = reQuot.exec(phrase)) {
      ts.push({type:'QUOT', word:ms[1], pos:pos + offset})
      pos = reQuot.lastIndex
      spAllowed = true
      continue
    }
    
    // ユーザーが正しくフレーズを書いた場合はここには来ない。
    // 大抵は、UOPの後にスペースを入れてしまった場合にここに来る。
    // なので、空文字列をトークンとして返す。Googleとの互換性を重視。
    ts.push({type:'LIT', word:'', pos:pos + offset})
    // posは変えない
    spAllowed = true
  }

  // EOP -- end of phrase
  ts.push({type:'EOP', word:'end-of-phrase', pos:pos + offset})
  return ts
}

const opMap = {
  '&': {prec:9,  alignL:true}, 
  '|': {prec:10, alignL:true}, 
  '~': {prec:1, alignL:true}  // concatenation i.e., a phrase "a b" will be ['~', a, b]
}

const precedes = (op1, op2) => {
  const op1spec = opMap[op1]
  const op2spec = opMap[op2]
  if (op1spec.prec == op2spec.prec) {
    return op1spec.alignL
  } else {
    return (op1spec.prec > op2spec.prec)
  }
}

// -> ast
// ast = [op, args, ...]  -- function application
//     | string  -- word
// where op = &, |, !, ~
const parse1 = (phrase, pos) => {
  const ts = tokenize(phrase, pos)
  let index = 0

  const term = () => {
    if (ts[index].type == 'LPAR') {
      const ast = comb()
      if (ts[index].type == 'RPAR') {
        index++
      }
      return ast
    } else if (ts[index].type == 'QUOT') {
      const t = ts[index++]
      return parse1(t.word, t.pos)
    } else if (ts[index].type == 'LIT') {
      const t = ts[index++]
      return t.word
    }
  }

  const fact = () => {
    if (ts[index].type == 'UOP') {
      const op = ts[index++].word
      if (op == 'site:') {
        // site:...の場合のみ特殊な操作
        const argIndex = index
        const arg = term()
        const arg2 = arg.replaceAll('.', ' ')
        const ast = parse1(arg2, argIndex)
        return ast
      } else if (op == '@' || op == '#') {
        const arg = term()
        return arg
      } else if (op == '-') {
        const arg = term()
        return [op, arg]
      } else {
        const arg = term()
        return null
      }
    } else {
      return term()
    }
  }

  const expr = () => {
    const stack = []
    stack.unshift(fact())
    while (ts[index].type == 'OP') {
      const op = ts[index++]
      if (! opMap.hasOwnProperty(op.word)) {
        throw new Error(`unexpected operator: ${op.word}`)
      }
      const rhs = fact()
      if (stack.length >= 3 && precedes(stack[1], op.word)) {
        const xb = stack.shift()
        const xop = stack.shift()
        const xa = stack.shift()
        stack.unshift([xop, xa, xb])
        stack.unshift(op.word)
        stack.unshift(rhs)
      } else {
        stack.unshift(op.word)
        stack.unshift(rhs)
      }
    }
    while (stack.length > 1) {
      const bx = stack.shift()
      const opx = stack.shift()
      const ax = stack.shift()
      stack.unshift([opx, ax, bx])
    }
    return stack[0]
  }

  const comb = () => {
    let first = expr()
    while (ts[index].type != 'EOP' && ts[index].type != 'RPAR') {
      const second = expr()
      first = ['~', first, second]
    }
    return first
  }
  
  const ast = comb()
  if (ts[index].type != 'EOP') {
    throw new Error(`unexpected token: ${JSON.stringify(ts[index])}`)
  }
  return ast
}

export const parse = (phrase) => {
  return parse1(phrase, 0)
}

// -> rsrcId[]
export const search = async (ast, lang, DB, withScore = false) => {
  // -> [scores:Record<rsrcId:key, score:number>, xlist:Record<rsrcId:key, any>, difficulty:number]
  const scale = (score) => (score - 1) * 0.2 + 1
  const evl = async (ast) => {
    if (ast === null) {
      return [{}, {}, 0]
    } else if (typeof ast == 'string') {
      const recs = await DB.sliceIndex(ast, lang)
      const scores = {}
      recs.forEach(rec => {
        scores[rec.rsrc] = scale(rec.score)
      })
      return [scores, {}, 1]

    } else if (ast[0] == '~') {
      const [scores1, xlist1, c1] = await evl(ast[1])
      const [scores2, xlist2, c2] = await evl(ast[2])
      const scores = {...scores1}
      const xlist = {...xlist1}
      for (let rsrc in scores2) {
        if (scores.hasOwnProperty(rsrc)) {
          scores[rsrc] += scores2[rsrc]
        } else {
          scores[rsrc] = scores2[rsrc]
        }
      }
      for (let rsrc in xlist2) {
        xlist[rsrc] = true
      }
      return [scores, xlist, c1 + c2]

    } else if (ast[0] == '-') {
      const [scores, xlist, c] = await evl(ast[1])
      return [{}, scores, 0]

    } else if (ast[0] == '&') {
      const [scores1, xlist1, c1] = await evl(ast[1])
      const [scores2, xlist2, c2] = await evl(ast[2])
      const scores = {}
      const xlist = {...xlist1}
      for (let rsrc in scores1) {
        if (scores2.hasOwnProperty(rsrc)) {
          scores[rsrc] = Math.max(scores1[rsrc], scores2[rsrc])
        }
      }
      for (let rsrc in xlist2) {
        xlist[rsrc] = true
      }
      return [scores, xlist, Math.max(c1, c2)]

    } else if (ast[0] == '|') {
      const [scores1, xlist1, c1] = await evl(ast[1])
      const [scores2, xlist2, c2] = await evl(ast[2])
      const scores = {...scores1}
      const xlist = {...xlist1}
      for (let rsrc in scores2) {
        if (scores.hasOwnProperty(rsrc)) {
          scores[rsrc] = Math.max(scores[rsrc], scores2[rsrc])
        } else {
          scores[rsrc] = scores2[rsrc]
        }
      }
      for (let rsrc in xlist2) {
        xlist[rsrc] = true
      }
      return [scores, xlist, Math.max(c1, c2)]

    } else {
      throw new Error(`unknown ast: ${JSON.stringify(ast)}`)
    }
  }

  const [scores, xlist, c] = await evl(ast)

  // excludes resources in xlist
  for (let rsrc in xlist) {
    delete scores[rsrc]
  }

  // define threshold
  const threshold = c * 0.65

  // sort by score 
  let rsrcs = []
  for (let rsrc in scores) {
    if (scores[rsrc] >= threshold) {
      rsrcs.push({rsrc, score:scores[rsrc]})
    }
  }
  rsrcs.sort((a, b) => a.score > b.score ? -1 : (a.score < b.score ? 1 : 0))

  if (withScore) {
    return rsrcs.slice(0, 10)
  } else {
    return rsrcs.slice(0, 10).map(a => a.rsrc)
  }
}

export const all = {parse, search}