
const MAX_DELAY = 30*60*60*1000  // 30 hours

/**
 * タスクにトランザクションという概念を導入する。
 * タスクのトランザクションとは、複数に分割されたタスクを全体として見る概念である。
 * たとえば、favの追加は、リソースの追加、APIサーバへの登録、キャプチャといった
 * 連続するタスクの集合であり、これらのタスクが1つのトランザクションを形成する。
 * 
 * タスクにはFGタスクとBGタスクがある。（フォアグラウンド・バックグウランド）
 * タスクの実行においては、FGタスクの処理が優先される。
 * FGタスクは一度トランザクションが開始されると最後まで処理される。従って、
 * FGタスクを安易に削除することはできない。
 * BGタスクはトランザクションの一部になってはならない。従って、BGタスクは
 * 削除しても他への影響は少ない。
 */

// -> task
export const createFavAdd = (urls, prior, after) => {
  return {
    id: null, 
    type: 'favAdd', 
    when: Date.now() + after, 
    trial: 0, 
    prior, 
    urls
  }
}

// -> task
export const createFavDel = (urls, rsrcIds, prior, after) => {
  return {
    id: null, 
    type: 'favDel', 
    when: Date.now() + after, 
    trial: 0, 
    prior, 
    urls, 
    rsrcIds
  }
}

// -> task
export const createCapture = (url, langs, prior, after) => {
  return {
    id: null, 
    type: 'capture', 
    when: Date.now() + after, 
    trial: 0, 
    prior, 
    url, 
    langs
  }
}

export const createFlip = (urls, batchName, batchSerial, prior, after) => {
  return {
    id: null, 
    type: 'flip', 
    when: Date.now() + after, 
    trial: 0, 
    prior, 
    urls, 
    batchName, 
    batchSerial
  }
}

// -> task
export const reschedule = (task) => {
  const newTask = {...task}
  newTask.id = null
  newTask.trial++
  // delays: 5s, 25s, 2m, 10m, 50m, 4.3h, 21.7h, 30h
  const delay = Math.min(Math.pow(5, newTask.trial), MAX_DELAY)
  newTask.when += delay * 1000
  return newTask
}

// -> void
export const putOff = (task) => {
  task.trial++
  // delays: 5s, 25s, 2m, 10m, 50m, 4.3h, 21.7h, 30h
  const delay = Math.min(Math.pow(5, task.trial), MAX_DELAY)
  task.when += delay * 1000
}

export const createBatch = (name) => {
  return {
    name, 
    serial: Date.now(), 
    totalSteps: 0, 
    currentSteps: 0, 
    numRejected: 0, 
    numErrors: 0
  }
}

export const incrTotalStep = (batch) => {
  batch.totalSteps++
}

export const incrCurrentStep = (batch) => {
  batch.currentSteps++
}

export const all = {createFavAdd, createFavDel, createCapture, createFlip, reschedule, putOff, createBatch, incrTotalStep, incrCurrentStep}
