
import {all as U} from '../utils.js';  // utilities

let baseUrl = null
let rsrcBaseUrl = null
let DB = null
let accessToken = null
let refleshToken = null
let L = null

const getHeader = (name, headers) => {
  const name1 = name.toLowerCase()
  for (let pair of headers.entries()) {
    if (pair[0].toLowerCase() == name1) {
      return pair[1]
    }
  }
  return null
}

const loadTokens = async () => {
  const data = await DB.findProp('credential')
  if (data) {
    accessToken = data.accessToken
    refleshToken = data.refleshToken
    L.log('token loaded', {accessToken})
  }
}

const storeTokens = async () => {
  await DB.storeProp('credential', {accessToken, refleshToken})
}

const responseToError = (resp) => {
  return new Error(`API Failure [${resp.status}] on ${resp.url}`)
}

// -> void
export const init = async (aBaseUrl, aRsrcBaseUrl, aDB, aL) => {
  baseUrl = aBaseUrl
  rsrcBaseUrl = aRsrcBaseUrl
  DB = aDB
  L = aL
  await loadTokens()

  if (! accessToken) {
    // init()関数はネットワーク要因ではエラーを出さない。もみ消す。
    // 各APIを呼び出すときに、必要に応じてregister()/reflesh()が走るので問題ない。
    try {
      await register()
    } catch (err) {
      L.error('on API.init()', L.flattenError(err))
    }
  }
}

const register = async () => {
  const options = {
    method: 'POST', 
    mode: 'cors', 
    headers: {
      'Content-Type': 'application/json'
    }, 
    body: JSON.stringify(null)
  }
  const resp = await fetch(baseUrl + '/user', options)
  if (! resp.ok) {
    throw responseToError(resp)
  }
  refleshToken = await resp.json()
  accessToken = getHeader('X-FAVALOT-ACCESS-TOKEN', resp.headers)
  L.log('tokens published', {accessToken, refleshToken})
  await storeTokens()
}

const reflesh = async () => {
  if (! refleshToken) {
    register()
    return
  }

  const options = {
    method: 'POST', 
    mode: 'cors', 
    headers: {
      'Content-Type': 'application/json'
    }, 
    body: JSON.stringify(refleshToken)
  }
  const resp = await fetch(baseUrl + '/user/reflesh', options)
  if (! resp.ok) {
    throw responseToError(resp)
  }
  accessToken = getHeader('X-FAVALOT-ACCESS-TOKEN', resp.headers)
  L.log('reflesh done', accessToken)
  await storeTokens()
}

const call = async (method, url, body) => {
  const options = {
    method, 
    mode: 'cors', 
    headers: {
      'X-FAVALOT-ACCESS-TOKEN': accessToken
    }
  }
  if (method == 'POST' || method == 'PUT' || method == 'PATCH') {
    options.headers['Content-Type'] = 'application/json'
    options.body = JSON.stringify(body)
  }

  let resp = await fetch(url, options)
  if (resp.status == 401) {
    await reflesh()
    options.headers['X-FAVALOT-ACCESS-TOKEN'] = accessToken
    resp = await fetch(url, options)
  }
  if (! resp.ok) {
    throw responseToError(resp)
  }
  const txt = await resp.text()
  L.log(`API.call(${resp.status}) ${method} ${url}`, {status:resp.status, body:txt})
  return JSON.parse(txt)
}

// -> void
export const addFav = async (urls, langs) => {
  const body = {urls, langs}
  await call('POST', baseUrl + '/favs/add', body)
}

// -> void
export const delFav = async (urls) => {
  const body = {urls}
  await call('POST', baseUrl + '/favs/delete', body)
}

// -> rsrc
// rsrc = {url:string, reprs:Record<key:string,repr:repr>, langs:<lang:string,key:string>, etag:string}
// repr = {status:int,mimetype:string,title:string,description:string,created:number,terms:Record<term:string,score:number>}
export const capture = async (url) => {
  const hash = await U.sha256(url)
  const prefix = hash.slice(0, 3)
  const eurl = prefix + U.rawUrlEncode(url).replaceAll('%', '%25')  // TODO: %だけでいいのか？
  
  const options = {
    method: 'GET', 
    mode: 'cors', 
    headers: {
      'Content-Type': 'application/json'
    }
  }
  const resp = await fetch(rsrcBaseUrl + eurl, options)
  if (! resp.ok) {
    throw responseToError(resp)
  }
  const txt = await resp.text()
  L.log(`RSRC.call(${resp.status}) GET ${url}`, {status:resp.status, body:txt})
  return JSON.parse(txt)
}

export const all = {init, addFav, delFav, capture}