
import {Dexie} from "../../node_modules/dexie/dist/dexie.mjs"
//import "./dexie.js"

let db = null
let currentSec = null
let serial = 1

// -> void
export const init = async (dbname) => {
  if (db) {
    throw new Error('DB.init called twice')
  }
  db = new Dexie(dbname, {indexedDB:self.indexedDB, IDBKeyRange:self.IDBKeyRange})
  db.version(1).stores({
    // indice goes here.
    // ローカル環境では、"++id"など自動採番は使えなかった。ので手動採番にする。generateId()参照
    rsrcs: "id, &url, pickAt", 
    indice: "id, *terms, rsrc", 
    fgtasks: "id, when", 
    bgtasks: "id, when", 
    props: "id, &name"
  })
}

// -> any  -- return-value of proc
export const transaction = async (mode, tables, proc) => {
  return await db.transaction(mode, tables, proc)
}

// DBを使わずにユニークなID（数値）を作る。
// 制限
// - 1秒あたり1000以上のIDは作れない
// - 1秒以内にプロセスが再起動された場合、IDが重複する危険性がある
const generateId = () => {
  const sec = Math.floor(Date.now() / 1000)
  if (sec == currentSec && serial >= 999) {
    throw new Error('too many id generation in 1 sec.')
  }
  if (sec != currentSec) {
    currentSec = sec
    serial = 1
  }
  return currentSec * 1000 + serial++
}

// -> void
export const storeRsrc = async (rsrc) => {
  if (! rsrc.id) {
    rsrc.id = generateId()
  }
  await db.rsrcs.put(rsrc)
}

// -> void
export const deleteRsrc = async (rsrcId) => {
  await db.rsrcs.delete(rsrcId)
}

export const listRsrcs = async () => {
  const result = await db.rsrcs.toArray()
  return result
}

// -> rsrc|null
export const findRsrc = async (rsrcId) => {
  const result = await db.rsrcs.get(+rsrcId)
  return result ? result : null
}

// -> rsrc|null
export const findRsrcByUrl = async (url) => {
  const result = await db.rsrcs.where('url').equals(url).first()
  return result ? result : null
}

// -> {rsrcs, total}
export const sliceRsrcs = async (offset, limit) => {
  const total = await db.rsrcs.count()
  const rsrcs = await db.rsrcs.offset(offset).limit(limit).toArray()
  return {rsrcs, total}
}

export const listTasks = async (prior) => {
  const table = prior ? "fgtasks" : "bgtasks"
  const result = await db.table(table).toArray()
  return result
}

// -> void
export const storeTask = async (task) => {
  const table = task.prior ? "fgtasks" : "bgtasks"
  if (! task.id) {
    task.id = generateId()
  }
  await db.table(table).put(task)
}

// -> void
export const deleteTask = async (taskId, prior) => {
  const table = prior ? "fgtasks" : "bgtasks"
  await db.table(table).delete(+taskId)
}

// -> task|null
export const pickTask = async (prior) => {
  const table = prior ? "fgtasks" : "bgtasks"
  const now = Date.now()
  const result = await db.table(table).orderBy("when").first()
  return (result && result.when <= now) ? result : null
}

// -> index[]
export const sliceIndex = async (term, lang) => {
  const indice = await db.indice.where('terms').equals([term, lang]).toArray()
  return indice
}

export const listIndice = async () => {
  const indice = await db.indice.toArray()
  return indice
}

export const sliceIndexByRsrc = async (rsrcId) => {
  const indice = await db.indice.where('rsrc').equals(+rsrcId).toArray()
  return indice
}

export const deleteIndice = async (rsrcId) => {
  await db.indice.where('rsrc').equals(+rsrcId).delete()
}

export const storeIndice = async (indice) => {
  for (let idx of indice) {
    if (! idx.id) {
      idx.id = generateId()
    }
  }
  await db.indice.bulkPut(indice, null)
}

const profileKeys = {
  token: 1
}

export const listProps = async () => {
  const props = await db.props.toArray()
  return props
}

export const findProp = async (name) => {
  const result = await db.props.where('name').equals(name).first()
  return result ? result.value : null
}

export const storeProp = async (name, data) => {
  let record = await db.props.where('name').equals(name).first()
  if (! record) {
    record = {id:generateId(), name, value:data}
  } else {
    record.value = data
  }
  await db.props.put(record)
}

export const deleteProp = async (name) => {
  await db.props.where('name').equals(name).delete()
}

export const clear = async () => {
  await db.rsrcs.clear()
  await db.indice.clear()
  await db.fgtasks.clear()
  await db.bgtasks.clear()
  await db.props.clear()
}

export const all = {init, transaction, storeRsrc, deleteRsrc, listRsrcs, findRsrc, findRsrcByUrl, sliceRsrcs, listTasks, storeTask, deleteTask, pickTask, sliceIndex, listIndice, sliceIndexByRsrc, deleteIndice, storeIndice, listProps, findProp, storeProp, deleteProp, clear}