

const genericErrorMessage = 'エラーが起きたため処理を中断しました。'

const download = (filename, blob) => {
  const download = document.createElement("a")
  download.download = filename
  download.href = window.URL.createObjectURL(blob)
  download.click()
  
  // no need to cleanup
}

/**
 * 
 * @param {boolean} going
 * @param {number} progress 
 */
const setExportProgress = (going, progress) => {
  if (! going) {
    document.getElementById('export-status').classList.add('hidden')
    document.querySelectorAll('.export-ui').forEach(el => {
      el.removeAttribute("disabled")
    })
  } else {
    ui("#export-progress-ui", progress)
    document.getElementById('export-progress-text').innerText = "" + progress + "%"
    document.getElementById('export-status').classList.remove('hidden')
    document.querySelectorAll('.export-ui').forEach(el => {
      el.setAttribute("disabled", true)
    })
  }
}

/**
 * 
 * @param {boolean} going 
 * @param {number|null} progress 
 */
const setImportProgress = (going, progress) => {
  if (! going) {
    document.getElementById('import-status').classList.add('hidden')
    document.querySelectorAll('.import-ui').forEach(el => {
      el.removeAttribute("disabled")
    })
  } else {
    if (progress == null) {
      document.getElementById('import-progress-ui').classList.add('indeterminate')
      document.getElementById('import-progress-text').innerText = ""
      document.getElementById('stop-import').setAttribute("disabled", true)
    } else {
      document.getElementById('import-progress-ui').classList.remove('indeterminate')
      ui("#import-progress-ui", progress)
      document.getElementById('import-progress-text').innerText = "" + progress + "%"
      document.getElementById('stop-import').removeAttribute('disabled')
    }
    document.getElementById('import-status').classList.remove('hidden')
    document.querySelectorAll('.import-ui').forEach(el => {
      el.setAttribute("disabled", true)
    })
  }
}

const exportLog = async () => {
  const message = {type:'dumplog'}
  const {ok, value} = await chrome.runtime.sendMessage(message)
  if (! ok) {
    window.alert(genericErrorMessage)
    return
  }

  download('favalotLog.json', new Blob([value], {type:'application/json'}))
}

const exportFavs = async () => {
  const hitsToText = (hits) => {
    return hits
      .map(hit => `"${hit.text.replace('"', '""')}","${hit.url}"`)
      .join("\r\n")
  }
  let page = 1
  const pages = []
  let amount = 0
  let lastPage = 0
  setExportProgress(true, 0)
  do {
    const message = {type:'favList', page}
    const {ok, value} = await chrome.runtime.sendMessage(message)
    if (! ok) {
      window.alert(genericErrorMessage)
      return
    }

    if (value.hits.length > 0) {
      amount += value.hits.length
      pages.push(hitsToText(value.hits))
      lastPage = value.lastPage
      setExportProgress(true, Math.floor(100 * page / lastPage))
    }
    page++
  } while (page <= lastPage)

  setExportProgress(true, 100)
  download('favalotBookmarks.csv', new Blob(pages, {type:'text/csv'}))
  setTimeout(() => {setExportProgress(false, 0)}, 5000)
}

async function* readLines(file) {
  const BLOCK_SIZE = 4096
  const eol = /\r?\n/
  let rest = ''
  let offset = 0
  do {
    const block = await file.slice(offset, offset + BLOCK_SIZE).text()
    const lines = (rest + block).split(eol)
    if (lines.length > 0) rest = lines.pop()
    for (let line of lines) {
      yield line
    }
    offset = offset + BLOCK_SIZE
  } while (offset < file.size)
  if (rest) {
    yield rest
  }
}

async function* readBookmarks() {
  const stack = []  // [nodes, index]
  let nodes = await chrome.bookmarks.getTree()
  let index = 0
  while (index < nodes.length || stack.length) {
    if (index >= nodes.length) {
      // no current node; then pop stack
      ([nodes, index] = stack.pop())
    }
    if (index < nodes.length) {
      // there current node
      const node = nodes[index]
      if (node.url) {
        yield node.url
        index++
      } else {
        stack.push([nodes, index + 1])
        nodes = node.children
        index = 0
      }
    }
  }
}

async function* cookUrls(itr) {
  const patn = /https?:\/\/[^,"']+/
  let rv = []
  while (true) {
    const {value, done} = await itr.next()
    if (value) {
      const result = value.match(patn)
      if (result) {
        rv.push(result[0])
        if (rv.length == 10) {
          yield rv
          rv = []
        }
      }
    }
    if (done) break
  }
  if (rv.length > 0) {
    yield rv
  }
}

const importFromCsv = async (ev) => {
  const inputEl = ev.currentTarget
  const file = inputEl.files[0]
  const itr = cookUrls(readLines(file))
  await importFromX(itr)
  inputEl.value = ""
}

const importFromBookmark = async () => {
  if (! window.confirm('ブラウザのブックマークをfavalotにインポートします。よろしいですか？')) return
  const itr = cookUrls(readBookmarks())
  await importFromX(itr)
}

const importFromX = async (itr) => {
  const name = 'import-batch'

  // add batch
  const {ok:ok0, value:batch} = await chrome.runtime.sendMessage({type:'batchAdd', name})
  if (! ok0) return
  setImportProgress(true, null)

  // charge tasks to the batch
  let succeeded = true
  while (true) {
    const {value:urls, done} = await itr.next()
    if (urls && urls.length > 0) {
      const message = {type:'batchPrepare', name, serial:batch.serial, urls}
      const {ok, value} = await chrome.runtime.sendMessage(message)
      if (! ok) {
        succeeded = false
        break
      }
    }
    if (done) break
  }

  // error handling
  if (! succeeded) {
    await chrome.runtime.sendMessage({type:'batchDel', name})
    window.alert(genericErrorMessage)
    return
  }

  // force ui sync
  await syncImportUi()
}

const syncImportUi = async () => {
  const name = 'import-batch'

  const {ok, value:currentBatch} = await chrome.runtime.sendMessage({type:'batchGet', name})
  if (! ok) return
  if (! currentBatch) return
  
  const progress = Math.floor(100 * currentBatch.currentSteps / currentBatch.totalSteps)
  setImportProgress(true, progress)
  
  const intervalId = setInterval(async () => {
    const {ok, value:batch} = await chrome.runtime.sendMessage({type:'batchGet', name})
    if (batch) {
      const progress = Math.floor(100 * batch.currentSteps / batch.totalSteps)
      setImportProgress(true, progress)
      if (progress == 100) {
        setTimeout(finishImport, 5000)
      }
    } else {
      setImportProgress(false, null)
      clearInterval(intervalId)
      await chrome.runtime.sendMessage({type:'batchDel', name})
    }
  }, 1000)

}

const finishImport = async () => {
  const name = 'import-batch'

  const {ok, value} = await chrome.runtime.sendMessage({type:'batchDel', name})
  setImportProgress(false, null)
}

const stopImport = async () => {
  const name = 'import-batch'

  if (window.confirm('ブックマークのインポートを中止します。よろしいですか？')) {
    const {ok, value} = await chrome.runtime.sendMessage({type:'batchDel', name})
    setImportProgress(false, null)
  }
}

addEventListener('load', () => {
  const dumplogBtn = document.getElementById('dumplog')
  dumplogBtn.addEventListener('click', exportLog)

  const exportBtn = document.getElementById('export')
  exportBtn.addEventListener('click', exportFavs)

  const impbmkBtn = document.getElementById('import-bookmark')
  impbmkBtn.addEventListener('click', importFromBookmark)

  const impcsvBtn = document.getElementById('import-csv')
  impcsvBtn.addEventListener('change', importFromCsv)

  const stopimpBtn = document.getElementById('stop-import')
  stopimpBtn.addEventListener('click', stopImport)

  syncImportUi()
})